/* This file is part of gacopyz.
   Copyright (C) 2009-2025 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <gacopyz_priv.h>

const char *gacopyz_stage_name[gacopyz_stage_max] = {
	"connect",	
	"helo",	
	"envfrom",	
	"envrcpt",	
	"data",
	"eom",
	"eoh"
};
