/* -*- buffer-read-only: t -*- vi: set ro: */
/* THIS FILE IS GENERATED AUTOMATICALLY.  PLEASE DO NOT EDIT. */

struct node_drv nodetab[] = {
    { NULL, NULL, NULL, NULL },
    { print_type_string, mark_type_string, code_type_string, NULL },
    { print_type_symbol, mark_type_symbol, code_type_symbol, NULL },
    { print_type_number, NULL, code_type_number, NULL },
    { print_type_if, mark_type_if, code_type_if, optimize_type_if },
    { print_type_bin, mark_type_bin, code_type_bin, optimize_type_bin },
    { print_type_un, mark_type_un, code_type_un, optimize_type_un },
    { print_type_result, mark_type_result, code_type_result, optimize_type_result },
    { print_type_header, mark_type_header, code_type_header, optimize_type_header },
    { print_type_builtin, mark_type_builtin, code_type_builtin, optimize_type_builtin },
    { print_type_concat, mark_type_concat, code_type_concat, optimize_type_concat },
    { print_type_variable, mark_type_variable, code_type_variable, NULL },
    { print_type_asgn, mark_type_asgn, code_type_asgn, optimize_type_asgn },
    { print_type_arg, NULL, code_type_arg, NULL },
    { print_type_argx, mark_type_argx, code_type_argx, optimize_type_argx },
    { print_type_argv, mark_type_argv, code_type_argv, optimize_type_argv },
    { print_type_vaptr, mark_type_vaptr, code_type_vaptr, optimize_type_vaptr },
    { print_type_regex, mark_type_regex, code_type_regex, NULL },
    { print_type_regcomp, mark_type_regcomp, code_type_regcomp, optimize_type_regcomp },
    { print_type_catch, mark_type_catch, code_type_catch, optimize_type_catch },
    { print_type_try, mark_type_try, code_type_try, optimize_type_try },
    { print_type_throw, mark_type_throw, code_type_throw, optimize_type_throw },
    { print_type_echo, mark_type_echo, code_type_echo, optimize_type_echo },
    { print_type_return, mark_type_return, code_type_return, optimize_type_return },
    { print_type_call, mark_type_call, code_type_call, optimize_type_call },
    { print_type_switch, mark_type_switch, code_type_switch, optimize_type_switch },
    { print_type_next, NULL, code_type_next, NULL },
    { print_type_break, NULL, code_type_break, NULL },
    { print_type_loop, mark_type_loop, code_type_loop, optimize_type_loop },
    { print_type_backref, NULL, code_type_backref, NULL },
    { print_type_cast, mark_type_cast, code_type_cast, optimize_type_cast },
    { NULL, mark_type_funcdecl, code_type_funcdecl, optimize_type_funcdecl },
    { NULL, mark_type_progdecl, code_type_progdecl, optimize_type_progdecl },
    { print_type_offset, mark_type_offset, code_type_offset, optimize_type_offset },
    { print_type_sedcomp, mark_type_sedcomp, code_type_sedcomp, optimize_type_sedcomp },
    { print_type_sed, mark_type_sed, code_type_sed, optimize_type_sed },
};
