/* -*- buffer-read-only: t -*- vi: set ro:
   THIS FILE IS GENERATED AUTOMATICALLY.  PLEASE DO NOT EDIT.
*/
#define REG_PC       0
#define REG_TOS      1
#define REG_TOH      2
#define REG_BASE     3
#define REG_REG      4
#define REG_MATCHSTR 5
