/* This file is part of Mailfromd.
   Copyright (C) 2005-2025 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <libmf.h>

void
close_fds_above(int fd)
{
	int i;

	for (i = getmaxfd(); i > fd; i--)
		close(i);
}

void
close_fds_except(fd_set *exfd)
{
	int i;
	
	for (i = getmaxfd(); i >= 0; i--)
		if (!(i < FD_SETSIZE && FD_ISSET(i, exfd)))
			close(i);
}
